import { TextLineStream } from "https://deno.land/std@0.224.0/streams/mod.ts"
import { writeAll } from "https://deno.land/std@0.224.0/io/write_all.ts"
import ollama, { Message } from "npm:ollama@0.5.15"
import { crayon } from "https://deno.land/x/crayon@3.3.3/mod.ts"

import { History } from "./history.ts"

const header = (name: string) => `
  ${crayon.yellow("KUB")}

Kub is a LLM running on TCP which includes all messages in the same context.
As Kub is having several discussions at once, he will know you as ${crayon.blue.bold(name)}
────────────────────────────────────────────────────────────────────────────────

`

export async function session(conn: Deno.TcpConn, name: string, history: History) {
	console.log("Opening session for", name, "at", conn.remoteAddr.hostname, conn.remoteAddr.port)
	conn.setNoDelay(true)
	await write_encoded(conn, header(name))
	await prompt(conn, name)
	const lines = conn.readable
		.pipeThrough(new TextDecoderStream())
		.pipeThrough(new TextLineStream())
	for await (const line of lines) {
		if (line === "") continue
		if (line.length > 10_000) return
		console.log("user", conn.remoteAddr.hostname, "prompted", line)
		if (line.startsWith("/nick ")) {
			name = line.slice("/nick ".length)
			console.log("Changed name of", conn.remoteAddr.hostname, "to", name)
			await prompt(conn, name)
			continue
		}
		await write_encoded(conn, `\n  ${crayon.yellow("kub")}\n\n`)
		const user_message = make_message(name, line)
		const response_parts = await get_response_parts(history, user_message)
		let content = "", role = "assistant"
		for await (const part of response_parts) {
			await write_encoded(conn, part.message.content)
			content += part.message.content
			role = part.message.role
		}
		await write_encoded(conn, "\n\n")
		await prompt(conn, name)
		await history.push(user_message)
		await history.push({ role, content })
	}
}

async function prompt(connection: Deno.TcpConn, name: string) {
	await write_encoded(connection, `  ${crayon.blue.bold(name)}\n\n> `)
}

async function write_encoded(connection: Deno.TcpConn, text: string) {
	const encoded = new TextEncoder().encode(text)
	await writeAll(connection, encoded)
}

async function get_response_parts(messages: History, user_message: { role: string; content: string }) {
	return await ollama.chat({
		model: "llama3.2",
		messages: request_messages(messages.messages, user_message),
		stream: true,
		options: {},
	})
}

function make_message(name: string, line: string) {
	const user_content = `(message from:${name}) ${line}`
	const user_message = { role: "user", content: user_content }
	return user_message
}

function request_messages(history_messages: Message[], user_message: Message) {
	return [
		{ role: "system", content: "You are 'Kub', a bot which talks to people on the internet. Be concise and helpful." },
		...history_messages,
		user_message,
	] as Message[]
}
